"""DryGASCON reference implementation"""

__version__ = '1.0.9'
__title__ = 'drysponge'
__description__ = 'DryGASCON reference implementation'
__long_description__ = """
DryGASCON reference implementation"""
__uri__ = 'https://github.com/sebastien-riou/DryGASCON'
__doc__ = __description__ + ' <' + __uri__ + '>'
__author__ = 'Sebastien Riou'
# For all support requests, please open a new issue on GitHub
__email__ = 'matic@nimp.co.uk'
__license__ = 'Apache 2.0'
__copyright__ = ''
